from enum import Enum
from uuid import UUID

from serenity_types.refdata.asset import Asset


class TokenAsset(Asset):
    """
    A tokenized asset.
    """

    token_issuance_id: UUID
    """
    A reference to the Exposure UUID for the underlying token issuance. This is to
    allow for the fact that a number of tokens have been issued on more than one
    blockchain, e.g. DAI on Ethereum and Solana in our model would be two different
    assets, tok.dai.ethereum and tok.dai.solana, both pointing to a single DAI
    exposure linked to MakerDAO.
    """

    network_id: UUID
    """
    The blockchain on which this particular asset has been tokenized.
    """


class PegMechanism(Enum):
    """
    The mechanism used to establish and maintain the peg of one asset's price ot another.
    """

    TOKENIZED_CASH = 'TOKENIZED_CASH'
    """
    Pegging to a currency reserve-backed holding, e.g. USDC and BUSD.
    """

    TOKENIZED_ASSET = 'TOKENIZED_ASSET'
    """
    Pegging to an asset reserve-backed holding, e.g. PAXG.
    """

    ALGOSTABLE = 'ALGOSTABLE'
    """
    Pegging via an algorithmic stabilization mechanism, e.g. FRAX.
    """

    COLLATERIZED_DEBT_POSITION = 'COLLATERIZED_DEBT_POSITION'
    """
    A tokenized CDP (Collateralized Debt Position), generally over-collateralized
    with cryptoassets, e.g. DAI.
    """


class PeggedTokenAsset(TokenAsset):
    """
    A token whose price aims to track the price of an Exposure.
    """

    referenced_exposure_id: UUID
    """
    The exposure that is being pegged to, e.g. USD (for USDC) XAU (for PAXG).
    """

    peg_mechanism: PegMechanism
    """
    How the peg is established and maintained, e.g. backed 100% by reserves
    of assets with an identical exposure as referenced_exposure_id, or algorithmically
    stabilized via an arbitrage mechanism, etc..
    """


class WrappedTokenAsset(TokenAsset):
    """
    A token that allows another token Exposure to be represented on a non-native
    network. Typically this is done by locking the asset in a smart contract or
    by a custodian holding the asset in reserve and then issuing a token. At any
    time the holder has a right to redeem the wrapped asset for the underlying
    asset one for one, though there is implicit smart contract / counterparty
    risk involved as this is not guaranteed.
    """

    wrapped_token_issuance_id: UUID
    """
    The token issuance that this asset is wrapping, e.g. BTC for WBTC.
    """

    origin_network_id: UUID
    """
    The original network that the wrapped asset sits on, e.g. bitcoin for WBTC,
    or ethereum or WETH. Note in the case of WETH on Ethereum as an asset, the
    origin_network_id and network_id are the same, because it's just an ERC20
    encapsulation of the native ETH token.
    """



from serenity_types.utils.serialization import CamelModel


class SymbolAuthority(CamelModel):
    """
    An authoritative source for asset symbols.
    """

    short_name: str
    """
    Short name for the authority or symbology; used in enums.
    """

    description: str
    """
    Human-readable description of this authority or symbology.
    """


class XRefSymbol(CamelModel):
    """
    A simple that can be used for cross-referencing a Serenity asset ID to another,
    external source's symbols.
    """

    authority: str
    """
    The short name for the owning authority.
    """

    symbol: str
    """
    The corresponding symbol.
    """



from datetime import date
from typing import List, Optional
from uuid import UUID

from serenity_types.portfolio.core import AssetPosition
from serenity_types.pricing.core import MarkTime
from serenity_types.utils.serialization import CamelModel


class AssetFactorExposureMatrixElement(CamelModel):
    """
    A single element from the matrix of asset factor exposures, for one day.
    """

    asset_id: UUID
    """
    The unique ID of the asset from the portfolio.
    """

    factor: str
    """
    The name of the factor from the risk model, e.g. Momentum.
    """

    value: float
    """
    Factor loading for the given factor-asset pair (betas from the regression).
    """


class AssetFactorIndexWeight(CamelModel):
    """
    A single entry in a factor portfolio index, with the asset's weight in the portfolio.
    """

    asset_id: UUID
    """
    The unique ID of the asset from the portfolio.
    """

    weight: float
    """
    signed weight (long or short) of this asset in the factor index; this is a fraction, e.g. 0.25 for 25%.
    For long-only they sum to 1.0 and for long-short they sum to 0.0 in the current methodology.
    """


class AssetMatrixElement(CamelModel):
    """
    A single cell in an asset-asset matrix, e.g. a covariance matrix.
    """

    asset_id1: UUID
    """
    First unique asset ID in the pair.
    """

    asset_id2: UUID
    """
    Second unique asset ID in the pair.
    """

    value: float
    """
    Matrix cell value, e.g. correlation.
    """


class AssetResidualValue(CamelModel):
    """
    A single entry in a list of asset residuals from the covariance matrix.
    """

    asset_id: UUID
    """
    Unique asset ID being referenced.
    """

    value: float
    """
    Asset residual covariance value.
    """


class Risk(CamelModel):
    """
    Common triplet for risk measures used across the API.
    """
    factor_risk: float
    """
    The volatility or variance explained by the factor risk model.
    """

    specific_risk: float
    """
    The volatility or variance not explained by the model.
    """

    total_risk: float
    """
    The total volatility or variance along the measured dimension.
    """


class AssetRisk(Risk):
    """
    This object supports the “Breakdown” view By Asset as well as the sector breakdowns.
    """

    asset_id: UUID
    """
    The unique asset ID for the asset break-down of risk.
    """

    sector_levels: List[str]
    """
    A list of sector level names that identifies a “path” to a particular level in the sector taxonomy,
    as defined above; however, in this instance, this is the fully-qualified path to this particular asset,
    i.e. the leaf level.
    """


class FactorExposureValue(CamelModel):
    """
    A decimal or base currency exposure to a risk factor.
    """

    factor_exposure: float
    """
    Asset-level factor loading.
    """

    factor_exposure_base_ccy: float
    """
    The factor exposure expressed in base currency terms.
    """


class FactorReturn(CamelModel):
    """
    A single daily return for a factor portfolio.
    """
    factor: str
    """
    Name of the factor from the risk model, e.g. Momentum.
    """
    close_date: date

    value: float
    """
    Daily return for this particular factor on the given close date.
    """


class FactorMatrixElement(CamelModel):
    """
    A single cell in a factor-factor matrix, e.g. factor covariance.
    """
    factor1: str
    """
    Name of the first factor from the risk model, e.g. Momentum
    """

    factor2: str
    """
    Name of the second factor from the risk model, e.g. Size.
    """

    value: float
    """
    Matrix cell value, e.g. a covariance or a correlation.
    """


class RiskAttributionRequest(CamelModel):
    """
    Request to run a risk attribution on the given portfolio.
    """

    as_of_date: Optional[date]
    """
    UTC date midnight; default to latest, as yyyy-mm-dd.
    """

    portfolio: List[AssetPosition]
    """
    Portfolio constituents to analyze.
    """

    model_config_id: Optional[UUID]
    """
    A reference to a factor risk model configuration from [ModelOps API, e.g. a specific parameterization
    of SFRMv1; if not specified then the system will select the most appropriate model.
    """

    sector_taxonomy_id: Optional[UUID]
    """
    References a taxonomy UUID from the Refdata API, specifically the getSectorTaxonomies call; defaults to DACS.
    """

    mark_time: Optional[MarkTime]
    """
    Time to use for close, defaults to MarkTime.NY_EOD.
    """

    base_currency_id: Optional[UUID]
    """
    Base currency for value reporting, defaults to USD.
    """

    strict: Optional[bool] = False
    """
    Should the request fail if there are any warnings?
    """


class SectorLevelRisk(Risk):
    """
    This object supports the “Breakdown” views By Sector & Asset and By Sector & Factor.
    The way to think about this one is the factor, specific & total risk for each supported risk
    contribution type, absolute and marginal at each level in the sector hierarchy gets one of these objects.
    """

    sector_levels: List[str]
    """
    A list of sector level names that identifies a “path” to a particular level in the sector taxonomy; as many of the
    risk values are non-additive, the output must provide every level explicitly, e.g.
    ['Sector A'], ['Sector A', 'Sub-sector A'] and ['Sector A', 'Sub-sector A', 'Industry A'] would
    all be in the by_sector array.
    """


class RiskBreakdown(CamelModel):
    """
    Collection of dimensions along which risk can be broken down.
    """

    by_sector: List[SectorLevelRisk]
    """
    The risk values at each level in the sector hierarchy for all of the sectors represented in the portfolio
    for the given risk measure, e.g. absolute vs. relative risk.
    """

    by_asset: List[AssetRisk]
    """
    The risk values for all of the assets in the portfolio for the given risk measure, e.g. absolute vs. relative risk.
    """


class SectorFactorExposure(CamelModel):
    """
    This object supports breaking down risk by sector and factor.
    """

    factor: str
    """
    Name of the factor from the risk model, e.g. Momentum.
    """

    sector_levels: List[str]
    """
    A list of sector level names that identifies a “path” to a particular level in the
    sector taxonomy, as defined above.
    """

    absolute_risk: float
    """
    Absolute contribution of this factor to the portfolio's overall volatility,
    i.e. the total amount of volatility attributable to this factor.
    """

    relative_risk: float
    """
    Relative contribution of this factor to the portfolio's overall volatility,
    i.e. the fraction of the volatility attributable to this factor.
    """

    marginal_risk: float
    """
    Marginal contribution of this factor to the portfolio's overall volatility,
    i.e. how much additional exposure to this factor would increase the volatility
    """

    factor_exposure: FactorExposureValue
    """
    The exposure at this hierarchy level to this factor.
    """


class TotalFactorRisk(CamelModel):
    """
    This object supports the Breakdown view By Factor, which shows the portfolio-level
    factor risk contributions and exposures.
    """

    factor: str
    """
    Name of the factor from the risk model, e.g. Momentum.
    """

    absolute_contribution: float
    """
    Absolute contribution of this factor to the portfolio's overall volatility,
    i.e. the total amount of volatility attributable to this factor
    """

    relative_contribution: float
    """
    Relative contribution of this factor to the portfolio's overall volatility,
    i.e. the fraction of the volatility attributable to this factor
    """

    marginal_contribution: float
    """
    Marginal contribution of this factor to the portfolio's overall volatility,
    i.e. how much would additional exposure to this factor increase the volatility
    """

    factor_exposure: FactorExposureValue
    """
    The exposure of the whole portfolio to this factor.
    """


class TotalRisk(CamelModel):
    """
    This object, returned at the top level, gives you a summary of the factor, specific and total risk expressed
    in both volatility and variance for the whole portfolio.
    """
    volatility: Risk
    """
    Factor, specific and total volatility of the portfolio.
    """

    variance: Risk
    """
    Factor, specific and total variance of the portfolio.
    """



from datetime import datetime
from uuid import UUID

from serenity_types.utils.serialization import CamelModel


class AssetMarkPrice(CamelModel):
    """
    A single asset price as of the chosen MarkTime.
    """

    asset_id: UUID
    """
    Unique ID of the referenced asset.
    """

    mark_time: datetime
    """
    The specific mark time used for latest price.
    """

    mark_price: float
    """
    The price at the mark time.
    """



from datetime import datetime
from typing import List, Optional
from uuid import UUID
from pydantic import validator

from serenity_types.pricing.derivatives.options.volsurface import (
    InterpolatedVolatilitySurface, VolModel, DiscountingMethod, ProjectionMethod
)
from serenity_types.pricing.derivatives.rates.yield_curve import InterpolatedYieldCurve
from serenity_types.refdata.options import OptionStyle, OptionType
from serenity_types.utils.serialization import CamelModel


class MarketDataOverride(CamelModel):
    """
    Helper type for representing replacements and bumps for market data inputs in pricing.
    """

    replacement: Optional[float]
    """
    Replacement value for the given market data point.
    """

    additive_bump: Optional[float]
    """
    A value (potentially negative) to add the observed value from stored or live market data.
    """

    @validator('replacement', always=True)
    def check_bump_or_replace_but_not_both(cls, replacement, values):
        if values.get('additive_bump') and replacement:
            raise ValueError("Please specify only one of 'replacement' or 'additive_bump'")
        return replacement


class YieldCurveOverride(CamelModel):
    """
    Helper for representing explicitly either override by UUID (load that YieldCurveDefinition
    from the database for the appropriate as_of_time) or by value. The client can also specify
    mutations to make to this input market data. In the case where you want to replace the
    rate entirely do not specify either yield_curve_id or yield_curve; in the case where you
    want to do a shift of the yield curve, specify either ID or curve and then an additive_bump.
    Everything null or both yield_curve_id and yield_curve specified yields validation errors.
    """

    yield_curve_id: Optional[UUID]
    """
    Optionally specifies a supported YieldCurveDefinition UUID from the database. Not every
    definition is accepted, e.g. you cannot pass in a CurveUsage.PROJECTION curve for discounting.
    """

    yield_curve: Optional[InterpolatedYieldCurve]
    """
    Optionally specifies a supported yield curve bootstrapped by the client or loaded separately. Not every
    definition is accepted, e.g. you cannot pass in a CurveUsage.PROJECTION curve for discounting.
    """

    rate_override: Optional[MarketDataOverride]
    """
    Optionally modifies the input data. Note properly you should not need to both provide a yield_curve
    and modify it, but in case clients want to play back our stored yield_curve via API without
    having to mutate the curve themselves, it could make sense.
    """

    @validator('yield_curve', always=True)
    def check_yield_curve_ids_or_yield_curve(cls, yield_curve, values):
        if values.get('yield_curve_id') and yield_curve:
            raise ValueError("Please specify only one of 'yield_curve_id' or 'yield_curve'")
        return yield_curve


class OptionValuation(CamelModel):
    """
    A collection of option economics and market data overrides used to describe a single option valuation
    requested from the service. This is intentionally meant to support both listed contract pricing and
    more general pricing of option economics. For market data, everything is defaulted, but the client can
    override or bump (shift) any of the inputs to get the exact pricing scenario desired.
    """

    option_valuation_id: str
    """
    Correlation ID to use for this requested option valuation. If pricing based on a listed contract
    with optionAssetId, by convention the unique ID or symbol of that option should be used.
    """

    qty: Optional[int] = 1
    """
    Number of option contracts; used when computing the spot notional of the option position. If you take
    the default every scaled value (e.g. spot_notional, delta_ccy) will be a unit notional value.
    """

    option_asset_id: Optional[UUID]
    """
    Look up all option economics based on the unique ID of a specific listed option contract.
    """

    underlier_asset_id: Optional[UUID]
    """
    Serenity asset identifier of the underlier, e.g. BTC (tok.btc.bitcoin). Not required if optionAssetId provided.
    Note we will be switching to the Exposure UUID instead in a future release (e.g. tok.btc), once the reference
    data is available.
    """

    strike: Optional[float]
    """
    Absolute value of the strike. Not required if optionAssetId provided. In future we may wish to support different
    StrikeType representations here, but some cases (like DELTA) are potentially trickier, so not for initial version.
    """

    expiry: Optional[datetime]
    """
    Expiration expressed in absolute terms as a date/time. Not required if optionAssetId provided
    """

    option_type: Optional[OptionType]
    """
    Whether we are pricing a PUT or CALL option.
    """

    option_style: Optional[OptionStyle] = OptionStyle.EUROPEAN
    """
    The variety of option being priced. Our pricer only supports EUROPEAN at this time, so defaults accordingly.
    """

    contract_size: Optional[float] = 1
    """
    For scaling purposes, the # of underlying per contract. Only used if option_asset_id is not set, otherwise it's
    loaded from the contract specification in the database.
    """

    implied_vol_override: Optional[MarketDataOverride]
    """
    Replace or modify the stored volatility surface's IV for this option.
    """

    spot_price_override: Optional[MarketDataOverride]
    """
    Replace or modify the stored or observed spot price used when pricing this option.
    """


class OptionValuationRequest(CamelModel):
    """
    A batch request to run one or more option valuations using a single model configuration and base
    set of curves and the vol surface. Reasonable defaults will be provided for any missing inputs, e.g.
    if you price a set of Deribit BTC options, the latest BTC volatility surface will be used along with
    the latest discounting curves for BTC and USD. Note that because the request only references a single
    volatility surface this means all included options must have the same underlier as the one in
    VolatilitySurfaceVersion.interpolated.definition.underlier_asset_id.
    """

    as_of_time: Optional[datetime]
    """
    The as-of time to use for loading all marketdata, surfaces, yield curves and refdata from the database.
    Defaulted to the latest up to this time.
    """

    model_config_id: Optional[UUID]
    """
    The specific derivatives analytics model configuration to load; this is used to drive defaults.
    """

    base_currency_id: Optional[UUID]
    """
    Base currency to use for expressing all notional values. Defaults to USD.
    """

    discounting_method: Optional[DiscountingMethod] = DiscountingMethod.SELF_DISCOUNTING
    """
    How to derive the discount rate: from the projection rate (self-discounting),
    or from pre-built discounting curves either provided in API or loaded from the system.
    """

    projection_method: Optional[ProjectionMethod]
    """
    How to derive the projection rate when in real-time mode: from live futures prices, or from a curve.
    The default depends on as_of_time. In the case of as_of_time being None the system runs in real-time
    pricing mode and uses ProjectionMode.FUTURES. When as_of_time is provided the system runs in historical
    pricing mode and defaults to ProjectionMode.CURVE. Setting both as_of_time and ProjectionMode.FUTURES
    will yield a validation error from the API.
    """

    vol_surface_id: Optional[UUID]
    """
    The optional unique ID of the surface to load, latest version as-of the as_of_time.
    """

    vol_surface: Optional[InterpolatedVolatilitySurface]
    """
    The optional client-provided volatility surface to use. If the client provides neither a VS ID
    nor their own volatility surface, the system will load the default for the underlying as-of the as_of_time.
    """

    discounting_curve_override: Optional[YieldCurveOverride]
    """
    Various forms of modifications to the discounting curve: choosing a variant in the database; passing
    in a complete curve by value; and/or replacing or shifting the extracted rate.
    """

    projection_curve_override: Optional[YieldCurveOverride]
    """
    Various forms of modifications to the projection curve: choosing a variant in the database; passing
    in a complete curve by value; and/or replacing or shifting the extracted rate.
    """

    vol_model: Optional[VolModel] = VolModel.SVI
    """
    The volatility model used for valuation purposes.
    """

    options: List[OptionValuation]
    """
    The full set of option valuations to run with the given market data inputs. The client may provide
    individual overrides or bumps for all inputs as part of each valuation object.
    """

    @validator('vol_surface', always=True)
    def check_vol_surface_id_or_vol_surface(cls, vol_surface, values):
        if values.get('vol_surface_id') and vol_surface:
            raise ValueError("Please specify only one of 'vol_surface_id' or 'vol_surface'")
        return vol_surface

    @validator('options', always=False)
    def check_non_empty_options_list(cls, options, values):
        if len(options) == 0:
            raise ValueError("Please provide at least one option in 'options'")
        return options


class OptionValuationResult(CamelModel):
    """
    The result of a series of option valuations based on the parameters in the OptionValuationRequest.
    Note that the basic calculation is just Black-Scholes, but if you provide additional information
    regarding the position scaling it will also provide position notional and greek exposures
    in base currency to allow bucketing of greeks and NAV calculations.
    """

    option_valuation_id: str
    """
    Correlation ID for the original OptionValuation.
    """

    vol_model: VolModel
    """
    The specific volatility model used; as SVI calibrations yield different greeks, this needs to be explicit.
    """

    pv: float
    """
    Present value (PV) a.k.a. theoretical price or theo.
    """

    iv: float
    """
    Implied volatility (IV)
    """

    spot_notional: float
    """
    The base currency notional of the position: number of contracts (qty) X  spot_price X contract_size.
    """

    spot_price: float
    """
    Input spot price for this valuation.
    """

    forward_price: float
    """
    Input forward price for this valuation.
    """

    projection_rate: float
    """
    The projection rate used when computing the forward.
    """

    discounting_rate: float
    """
    The discounting rate used when computing the forward; equal to projection_rate with SELF_DISCOUNTING.
    """

    delta: float
    """
    Greek output: delta, the option's sensitivity to spot changes.
    """

    delta_qty: float
    """
    Delta X qty X contract_size, the delta exposure expressed in qty of underlying.
    """

    delta_ccy: float
    """
    Delta X value, a.k.a. the partial derivative of position value with respect to spot,
    expressed in base currency
    """

    gamma: float
    """
    Greek output: gamma, the delta's sensitivity to spot changes.
    """

    gamma_ccy: float
    """
    Gamma X value^2, a.k.a. the second derivative of position value with respect to spot,
    expressed in base currency.
    """

    vega: float
    """
    Greek output: vega, the option's sensitivity to volatility changes.
    """

    vega_ccy: float
    """
    Partial derivative of the position value of the contract with respect to vega X 1%, expressed in base currency.
    """

    rho: float
    """
    Greek output: rho, the delta's sensitivity to interest rate changes.
    """

    rho_ccy: float
    """
    Partial derivative of the spot notional value of the contract with respect to rho X 1bp, expressed in base currency.
    """

    theta: float
    """
    Greek output: theta, the delta's sensitivity to time decay.
    """

    theta_ccy: float
    """
    Partial derivative of the spot notional value of the contract with respect to theta X 1 day,
    expressed in base currency.
    """



from enum import Enum
from typing import Optional
from uuid import UUID

from serenity_types.utils.serialization import CamelModel


class ExposureType(Enum):
    COUNTERPARTY = 'COUNTERPARTY'
    """
    Bilateral exposure to a particular counterparty.
    """

    TOKEN_ISSUANCE = 'TOKEN_ISSUANCE'
    """
    Exposure to an issuance of a token.
    """

    DEBT_ISSUANCE = 'DEBT_ISSUANCE'
    """
    Exposure to an issuance of debt by a sovereign or corporate entity.
    """

    EQUITY_ISSUANCE = 'EQUITY_ISSUANCE'
    """
    Exposure to an issuance of equity by a corporate entity, whether public or private.
    """

    FIAT_ISSUANCE = 'FIAT_ISSUANCE'
    """
    Exposure to a sovereign, supranational central bank (ECB) or multinational
    organization (the IMF, for SDR).
    """


class Exposure(CamelModel):
    """
    Generic exposure to some bundle of financial risks. In multiple contexts we sometimes
    want to refer not to a specific tradable instrument but instead to the exposure. For
    instance, a centralized exchange listing DAI-USDT is not going to guarantee for you
    that the DAI tokens are specifically those issued on the Ethereum blockchain. In Serenity,
    the latter is an Asset, while the general idea of DAI, which entails exposure to everything
    issued by MakerDAO as a DAI token on any blockchain, is an Exposure. This also lets us
    express things like USDT is pegged to USD, but that doesn't necessarily mean Tether holds
    USD currency positions; it aims to track the price of the dollar. Tether is a token Exposure
    which is pegged to a fiat Exposure, USD, which may be tokenized as an asset on Ethereum
    and then held in a wallet. Considering all of this more generally, when we assess the risks
    in a portfolio, we must consider that, say, USDT on Binance is just a SQL record; it is,
    at the end of the day, a claim on Binance to get that exposure which may or may not be
    honored, just as USDT is a claim on USD exposure which is not risk-free. Even a tokenized
    USD deposit embeds a dual counterparty risk: to the smart contract, and to the bank -- yet
    we still want to express this idea that it's holding dollars and this is in theory the
    same as what you could by redeeming USDT for USD.
    """

    exposure_id: UUID
    """
    Unique identifier for this exposure.
    """

    exposure_type: ExposureType
    """
    Category of exposure.
    """

    party_org_id: Optional[UUID]
    """
    Where relevant, the legal entity / party that creates this particular
    exposure, e.g the token issuer (e.g. Circle, MakerDAO), or in the case of a counterparty
    exposure to an OTC desk, the DEALER (e.g. Galaxy Digital). In the case of
    a fiat currency exposure, the party is the soverign or, in the case of EUR, the ECB.
    """

    symbol: str
    """
    Serenity's unique symbol for this exposure, e.g. tok.usdc or ccy.usd.
    """

    display_name: str
    """
    Human-readable name for this exposure, e.g. Ethereum or U.S. Dollar.
    """

    icon_uri: Optional[str]
    """
    Optional path for loading an icon representing this exposure; typically a URL.
    """



from enum import Enum

from serenity_types.refdata.derivatives import Expiry, ListedDerivative


class OptionStyle(Enum):
    """
    Variety of option supported. Right now we only support vanilla European options.
    """

    AMERICAN = 'AMERICAN'
    """
    An option that can be exercised at any time up to the expiry date.
    """

    EUROPEAN = 'EUROPEAN'
    """
    An option that can only be exercised on the expiry date.
    """


class OptionType(Enum):
    """
    Put/call flag for options.
    """

    PUT = 'PUT'
    """
    The right but not the obligation to sell the underlier at the strike price on exercise.
    """

    CALL = 'CALL'
    """
    The right but not the obligation to buy the underlier at the strike price on exercise.
    """


class ListedOption(ListedDerivative):
    """
    An exchange-listed vanilla option.
    """

    option_type: OptionType
    """
    Whether this is a put or call contract.
    """

    option_style: OptionStyle
    """
    Exercise type for this option, e.g. European exercise.
    """

    strike_price: float
    """
    Strike price for the option contract, e.g. 20000.
    """

    expiry: Expiry
    """
    Expiration date and time for this option.
    """



from datetime import datetime
from enum import Enum
from typing import Dict, List
from uuid import UUID

from serenity_types.utils.serialization import CamelModel


class VolModel(Enum):
    """
    Currently supported volatility models.
    """

    SVI = "SVI"
    """
    Stochastic volatility (SVI) calibrated volatility model.
    """

    BLACK_SCHOLES = "BLACK_SCHOLES"
    """
    Classic Black-Scholes volatility model.
    """


class StrikeType(Enum):
    """
    Currently supported strike representations.
    """

    ABSOLUTE = "ABSOLUTE"
    """
    Absolute value of strike, e.g. the 20000 option.
    """

    LOG_MONEYNESS = "LOG_MONEYNESS"
    """
    Relative value of strike vs. current spot with log transformation: log(strike / spot).
    """


class DiscountingMethod(Enum):
    """
    The strategy to use for deriving the discount rate, i.e. the assumed interest rate to
    use when discounting prices to present.
    """

    CURVE = "CURVE"
    """
    Extract a discount factor from discount curves, either provided via API or loaded from database.
    """

    SELF_DISCOUNTING = "SELF_DISCOUNTING"
    """
    Use the base asset's projection rate as the discount rate instead of a discount curve.
    This is the default, and the option you should choose if you want to reproduce the Deribit
    forward pricing model.
    """


class ProjectionMethod(Enum):
    """
    The strategy to use for deriving the projection rates, i.e. the forward interest rates.
    """

    CURVE = "CURVE"
    """
    Projection rate is extracted from a projection curve, either provided via API or loaded from database.
    This option is respected in both real-time pricing and historical pricing modes, though in real-time
    the curve version loaded is always the very latest. Select CURVE if you want a more stable forward.
    """

    FUTURES = "FUTURES"
    """
    Projection rate is snapped from the corresponding Deribit futures prices; this option is only
    supported in real-time pricing mode. Select FUTURES if in real-time you want to reproduce
    Deribit forward pricing model and incorporate up-to-the-moment market view on the forward.
    """


class VolatilitySurfaceDefinition(CamelModel):
    """
    A uniquely-identified set of VS parameters for fitting a VolatilitySurface.
    """

    vol_surface_id: UUID
    """
    Unique ID for this volatility surface's collection of attributes; note that surfaces
    are re-fitted hourly, and so there are going to be many versions over time.
    """

    vol_model: VolModel
    """
    Volatility model used for this surface.
    """

    strike_type: StrikeType
    """
    Strike representation used for this surface, e.g. ABSOLUTE or LOG_MONEYNESS.
    """

    underlier_asset_id: UUID
    """
    The linked asset for this surface, e.g. for a Bitcoin volatility surface, this is BTC (tok.btc.bitcoin).
    Note we will be switching to the Exposure UUID instead in a future release (e.g. tok.btc), once the
    reference data is available.
    """

    display_name: str
    """
    Human-readable description of this curve, e.g. Deribit BTC (SVI, ABSOLUTE)
    """


class VolatilitySurfaceAvailability(CamelModel):
    """
    Information about version availability for a given volsurface definition.
    """

    definition: VolatilitySurfaceDefinition
    """
    Description of the particular volsurface parameters that are available to load.
    """

    as_of_times: List[datetime]
    """
    The list of all available as_of_times in the requested window.
    """


class VolPoint(CamelModel):
    """
    An individual IV input point.
    """

    option_asset_id: UUID
    """
    The specific option that was used for vol fitting purposes.
    """

    time_to_expiry: float
    """
    The time to expiry for this point, expressed as a year fraction.
    """

    strike_value: float
    """
    Value of strike for this point, unit defined by StrikeType.
    """

    mark_price: float
    """
    The observed option premium used as input to the IV calculation.
    """

    projection_rate: float
    """
    The projection rate used when computing the forward.
    """

    discounting_rate: float
    """
    The discounting rate used when computing the forward; equal to projection_rate with SELF_DISCOUNTING,
    which is currently the default for our IV calculations.
    """

    forward_price: float
    """
    The computed forward price that went into the IV calculation.
    """

    iv: float
    """
    The computed implied volatility (IV) that corresponds to the given mark_price and other inputs.
    """


class RawVolatilitySurface(CamelModel):
    strike_type: StrikeType
    """
    Strike representation used for this surface, e.g. ABSOLUTE or LOG_MONEYNESS.
    """

    spot_price: float
    """
    The observed spot price that went into the IV calculations.
    """

    vol_points: List[VolPoint]
    """
    The discrete IV points available for fitting as a volatility surface.
    """


class InterpolatedVolatilitySurface(CamelModel):
    """
    A calibrated volatility surface with a dense grid of fitted vols. Each array
    is of equal length and corresponds to (x, y, z) for the mesh.
    """

    definition: VolatilitySurfaceDefinition
    """
    The unique set of parameters used to calibrate / fit this surface.
    """

    strikes: List[float]
    """
    All strikes expressed as log-money values, the x-axis in the mesh.
    """

    time_to_expiries: List[float]
    """
    All times to expiry expressed as year fractions, the y-axis in the mesh.
    """

    vols: List[float]
    """
    All fitted vols, the z-axis in the mesh.
    """

    input_params: Dict[str, object]
    """
    Informational set of input parameters, e.g. yield curves used for the forward. May be empty
    and keys will depend on the configuration, e.g. DiscountingType.
    """

    calibration_params: Dict[float, Dict[str, float]]
    """
    Informational set of calibration parameters, e.g. the SVI parameters. Keying is time_to_expiry
    expressed in year fractions to parameter set, where the parameter keys are VolModel-specific.
    """


class VolatilitySurfaceVersion(CamelModel):
    """
    A single version of a fitted volatility surface, with both the raw and interpolated content.
    """

    raw: RawVolatilitySurface
    """
    The raw volatility surface inputs.
    """

    interpolated: InterpolatedVolatilitySurface
    """
    The interpolated volatility surface.
    """

    as_of_time: datetime
    """
    The time window, generally top of the hour, for which we have fitted the volatility surface; latest prices
    as of this time are used as input to the surface calibration.
    """

    build_time: datetime
    """
    The actual time of the build; due to DQ or system issues this might be different from as_of_time.
    """



from typing import List

from serenity_types.utils.serialization import CamelModel


class LookupRow(CamelModel):
    """
    A single mapping from a server-side enum to supporting metadata for Serenity UX.
    """

    key: str
    """
    The name of the enum in code, e.g. FOO_BAR.
    """

    display_name: str
    """
    A human-readable short name for the name, e.g. Foobarity.
    """

    tooltip: str
    """
    A human-readable description of this particular key suitable for tooltips, e.g. “A statistical measure of
    fundamental foobariosity; may be purple.”
    """


class LookupTable(CamelModel):
    """
    A simple collection of enum metadata to be used by the Serenity front-end to map API codes to human-readable
    strings, tooltips, etc..
    """

    table_name: str
    """
    Short code for the lookup table, e.g. ASSET_CLASSES.
    """

    rows: List[LookupRow]
    """
    All rows in this particular lookup table.
    """



from datetime import date
from typing import Dict
from uuid import UUID

from serenity_types.utils.serialization import CamelModel


class Model(CamelModel):
    """
    A risk model or other model built into Serenity. In general Models correspond to a single
    implementation in code, potentially with multiple parameterizations represented by several
    ModelConfiguration entries.
    """

    model_id: UUID
    """
    Unique ID for this particular model.
    """

    model_class_id: UUID
    """
    Unique ID for the parent ModelClass.
    """

    short_name: str
    """
    A structured name suitable for lookup keys, like risk.factor.regression.slm.
    """

    display_name: str
    """
    A human-readable name suitable for labels, like Serenity Factor Risk Model.
    """

    description: str
    """
    A longer, human-readable description suitable for tooltips and docs.
    """


class ModelClass(CamelModel):
    """
    A group of related models, like all VaR models.
    """

    model_class_id: UUID
    """
    Unique ID for this class of models.
    """

    short_name: str
    """
    A structured name suitable for lookup keys, like risk.market.
    """

    display_name: str
    """
    A human-readable name suitable for labels, like Market Risk.
    """

    description: str
    """
    A longer, human-readable description suitable for tooltips and docs.
    """


class ModelConfiguration(CamelModel):
    """
    A parameterizxation of a Model, e.g. the short, medium and long time horizon variations
    of a factor risk model.
    """

    model_config_id: UUID
    """
    Unique ID for this specific model version and parameterization.
    """

    model_id: UUID
    """
    Unique ID for the parent Model.
    """

    short_name: str
    """
    A structured name suitable for lookup keys, like risk.var.parametric.normal.
    """

    display_name: str
    """
    A human-readable name suitable for labels, like Parametric VaR Model (Normal).
    """

    description: str
    """
    A longer, human-readable description suitable for tooltips and docs.
    """

    version: str
    """
    Version number, by convention following SemVer conventions.
    """

    deployment_date: date
    """
    UTC date when the config was last deployed.
    """

    training_date: date
    """
    UTC date for as-of date of training data used.
    """

    code_version: str
    """
    A GitHub commit ID, tag or other means to uniquely identify the code.
    """

    data_parameters: Dict[str, object]
    """
    A set of name-value pairs characterizing input data.
    """

    model_parameters: Dict[str, object]
    """
    A set of name-value pairs parameterizing the model.
    """


class ModelConfigurationSummary(CamelModel):
    """
    Summary of a particular available model configuration, suitable for external use.
    """

    model_config_id: UUID
    """
    Unique ID for this specific model version and parameterization.
    """

    model_id: UUID
    """
    Unique ID for the parent Model.
    """

    short_name: str
    """
    A structured name suitable for lookup keys, like risk.var.parametric.normal.
    """

    display_name: str
    """
    A human-readable name suitable for labels, like Parametric VaR Model (Normal).
    """

    description: str
    """
    A longer, human-readable description suitable for tooltips and docs.
    """



from serenity_types.refdata.asset import Asset


class Currency(Asset):
    """
    An asset representing a fiat currency like the dollar, euro or yen. Note this will
    need to be extended to link to a fiat_issuance_id for the Exposure UUID once we
    have distinct fiat exposure reference data.
    """

    iso_currency_code: str
    """
    The ISO currency code, e.g. USD or EUR.
    """



from typing import List
from uuid import UUID


class AssetPosition:
    """
    A simple representation of holding a certain amount of a given asset.
    This is going to be replaced with a much more flexible Portfolio
    representation in Q1'23.
    """

    asset_id: UUID
    """
    Unique identifier of the asset from Serenity's asset master database.
    """

    quantity: float
    """
    The number of tokens, shares, contracts, etc. held in this position.
    If positive this indicates a long position; if negative, a short one.
    """


class SimplePortfolio:
    """
    A simple portfolio representation that just maps the positions to
    positive or negative quantities for long and short. There is no
    history, detail on custody or any other context.
    """

    portfolio_id: UUID
    """
    Unique ID; in the initial implementation this is assigned locally in
    a client installation-hosted database.
    """

    base_currency_id: UUID
    """
    Asset ID of the base currency for this portfolio.
    """

    portfolio_name: str
    """
    Descriptive name for this portfolio, for display only.
    """

    portfolio_manager: str
    """
    In the initial implementation, a text field for the PM;
    eventually will link to user ID in the database.
    """

    positions: List[AssetPosition]
    """
    List of positions in the portfolio.
    """



from datetime import date, datetime
from enum import Enum
from typing import List, Optional, Union
from uuid import UUID

from serenity_types.portfolio.core import SimplePortfolio
from serenity_types.pricing.core import PricingContext
from serenity_types.utils.serialization import CamelModel


class ShockPnL(CamelModel):
    """
    Value object with information about a particular dimension's P&L impact from a single shock.
    """

    shock_id: UUID
    """
    Unique ID for the shock from the original ShockDefinition.
    """

    shock_pnl: float
    """
    The PnL, in base currency, from the shock, e.g. the portfolio would have lost an
    additional $2K if this shock had been applied.
    """


class PnL(CamelModel):
    """
    Value object with information about a particular dimension's P&L impact from all shocks.
    """

    base_pnl: float
    """
    The PnL, in base currency, over the period (default 1 day) absent all the shocks, e.g.
    portfolio lost $12K in the last 24 hours.
    """

    total_shock_pnl: float
    """
    The PnL, in base currency, from all shocks combined, e.g. the portfolio would have lost an additional
    $50K if all shocks had been applied.
    """

    pnl_by_shock: List[ShockPnL]
    """
    The PnL, in base currency, from applying each shock individually, e.g. the portfolio would have lost
    $4K from shocking ETH down 10% and $35K from shocking BTC down 10%.
    """


class AssetPnL(CamelModel):
    """
    A single entry representing an asset-level impact of this scenario's shocks.
    It supports both by-asset and by-sector views in Serenity UX.
    """

    asset_id: UUID
    """
    Unique asset identifier from the Asset Master.
    """

    sector_levels: List[str]
    """
    A list of sector level names that identifies a “path” to a particular
    level in the sector taxonomy, as defined above; note we only have the
    leaf level here; we don't need to show PnL at each level in the sector
    hierarchy above the assets, UX can do the aggregation since $ values
    are additive here.
    """

    asset_pnl: PnL
    """
    P&L impact in total and by shock for this asset.
    """


class FactorPnL(CamelModel):
    """
    A single entry representing a factor-level impact of this scenario's shocks.
    """
    pass

    factor: str
    """
    Factor name, e.g. MOMENTUM.
    """

    factor_exposure_base_ccy: float
    """
    Total factor exposure in the portfolio, in base currency.
    """

    factor_pnl: PnL
    """
    P&L impact in total and by shock for this factor.
    """


class ScenarioSource(Enum):
    """
    Enum classifying scenarios by whether they are user- or system-defined.
    """

    CUSTOM = 'CUSTOM'
    """
    A user-defined scenario.
    """

    PREDEFINED = 'PREDEFINED'
    """
    A canned scenario provided by the platform; typically these represent
    fitted historical events like the 2020 COVID Crash.
    """


class ShockTo(Enum):
    """
    Enum classifying what is being shocked, e.g. asset returns or factor returns.
    """

    ASSET = "ASSET"
    """
    Shock applied additively to the asset's returns at the close, per MarkTime.
    """

    FACTOR = "FACTOR"
    """
    Shock applied additively to the factor's returns at the close, per MarkTime.
    """


class Shock(CamelModel):
    """
    A discrete market shock that is being simulated in a scenario.
    """

    shock_id: UUID
    """
    Unique ID for the shock; this can be generated and stored with the ShockDefinition (and thus unchanging)
    or if the scenario is run on-the-fly, the UUID should be generate at request time so there is a clear
    linkage between Shock and ShockPnL.
    """

    target_type: ShockTo
    """
    What to shock: asset returns or factor returns?
    """

    target: Union[UUID, str]
    """
    UUID for asset ID, or name of the factor to shock.
    """

    magnitude: float
    """
    Magnitude of the shock, + / - for up or down shock actions; note this is expressed as a fraction, not a percentage.
    """


class ScenarioDefinition(CamelModel):
    """
    A complete definition of a stress test to perform for scenario analysis.
    """

    scenario_id: Optional[UUID]
    """
    Unique ID for this scenario, optional when creating or if we are dealing with a transient scenario to run.
    """

    scenario_version: Optional[int]
    """
    Monotonically increasing version number, optional when creating or if we are dealing
    with a transient scenario to run.
    """

    source: ScenarioSource
    """
    Whether this shock is user- or system-defined.
    """

    name: str
    """
    A descriptive name for this scenario.
    """

    shocks: List[Shock]
    """
    The discrete shocks to apply as part of this scenario.
    """

    deleted: Optional[bool] = False
    """
    If set, this definition was logically deleted in the database and is not current.
    """

    last_updated: Optional[datetime]
    """
    Last update timestamp, in UTC.
    """

    last_updated_by: Optional[str]
    """
    Last update user.
    """


class ScenarioCloneRequest(CamelModel):
    """
    A request to clone an existing scenario; typically used to allow the client
    to customize a canned scenario.
    """

    scenario_id: UUID
    """
    Unique ID of the custom or predefined scenario to clone.
    """

    scenario_name: str
    """
    Name to use post-cloning to avoid ambiguity.
    """


class ScenarioRequest(CamelModel):
    """
    A request to run a scenario against a portfolio. This scenario may be stored or transient.
    """

    scenario_id: Optional[UUID]
    """
    Unique ID of the scenario to run; if not provided, user must provide a transient scenario instead.
    """

    scenario: Optional[ScenarioDefinition]
    """
    Transient scenario to run; if not provided, user must provide a scenario ID instead.
    """

    portfolio: SimplePortfolio
    """
    Portfolio to run the scenario on.
    """

    pricing_context: PricingContext
    """
    Common model inputs, e.g. mark time.
    """

    sector_taxonomy_id: Optional[UUID]
    """
    References a taxonomy UUID from the Refdata API, specifically the getSectorTaxonomies call.
    """

    model_config_id: UUID
    """
    Factor model configuration ID to use for risk purposes.
    """

    start_date: date
    """
    Start of the shock period; in the instantaneous case, same as end_date.
    """

    end_date: date
    """
    End of the shock period; in the instantaneous case, same as start_date.
    """


class ScenarioResult(CamelModel):
    """
    ScenarioResult provides the portfolio base & shock P&L and the breakdowns by asset, sector and by factor.
    """

    portfolio_pnl: PnL
    """
    The impact to the portfolio broken out by shocks.
    """

    asset_pnl: List[AssetPnL]
    """
    Report on P&L impact by asset & sector.
    """

    factor_pnl: List[FactorPnL]
    """
    Report on P&L impact by risk factor in the provided factor risk model.
    """


class RunStatus(Enum):
    """
    Results of the scenario run at a high level.
    """

    RUNNING = 'RUNNING'
    """
    Job sent to the scenario engine and running, may be async / long-running.
    """

    COMPLETED = 'COMPLETED'
    """
    Job completed with no errors.
    """

    FAILED = 'FAILED'
    """
    Job failed with an error.
    """


class ScenarioRun(CamelModel):
    """
    A summary of the run. Note that to get the detail you need to call getScenarioRun with the runId;
    this just gives you the status, the base PnL absent all the shocks and the total shock PnL.
    """

    run_id: UUID
    """
    Unique ID of this run.
    """

    scenario_request: ScenarioRequest
    """
    The request that was run previously.
    """

    status: RunStatus
    """
    Current state of the run.
    """

    base_pnl: float
    """
    The PnL, in base currency, over the period (default 1 day) absent all the shocks,
    e.g. portfolio lost $12K in the last 24 hours.
    """

    total_shock_pnl: float
    """
    the PnL, in base currency, from all shocks combined, e.g. the portfolio would have
    lost an additional $50K if all shocks had been applied.
    """

    start_datetime: datetime
    """
    UTC datetime of the run start.
    """

    end_datetime: Optional[datetime]
    """
    UTC datetime of the run completion, if COMPLETED.
    """



from datetime import date, datetime
from enum import Enum
from typing import List, Optional
from uuid import UUID

from serenity_types.utils.serialization import CamelModel


class CurveUsage(Enum):
    """
    Intended usage of this curve, e.g. for discounting or for projection purposes.
    """

    DISCOUNTING = "DISCOUNTING"
    """
    Curve points for discounting future cashflows.
    """

    PROJECTION = "PROJECTION"
    """
    Curve points indicating the market forward view.
    """


class RateSourceType(Enum):
    """
    Sources of rates & discount factors. In the most general case
    the yield curve can be built from multiple imports, so we
    tag each CurvePoint that is the input to the interpolated YC
    with the particular source, e.g. we might have an 8H rate
    at the short end from perpetual future funding rates and
    implied forwards backed out from calendar spreads.
    """

    FUTURE_PX = "FUTURE_PX"
    """
    Implied forward backed out from observed future prices.
    """

    OPTION_PX = "OPTION_PX"
    """
    Implied forward backed out from observed option prices & spreads.
    """

    FIXING = "FIXING"
    """
    Observed traditional rate fixings.
    """

    FUNDING_RATE = "FUNDING_RATE"
    """
    Observed exchange perpetual future funding rate.
    """

    LENDING_RATE = "LENDING_RATE"
    """
    Observed CeFi / OTC or DeFi lending rate.
    """

    STAKING_RATE = "STAKING_RATE"
    """
    Observed proof-of-stake protocol staking rate.
    """


class InterpolationMethod(Enum):
    """
    Specific interpolation method used. Currently only supports flat-forward.
    """
    FLAT_FWD = "FLAT_FWD"


class CurvePoint(CamelModel):
    """
    A discrete input point on the curve, with all the metadata describing
    what is being provided and its source to help reproduce the results.
    """

    tenor: Optional[str]
    """
    A relative date code, e.g. 1Y or 3M.
    """

    pillar_date: Optional[date]
    """
    The specific forward date for the given rate and DF, e.g. the 1W point for today would be the next
    business day a week ahead.
    """

    duration: Optional[float]
    """
    The duration for this point, expressed as a year fraction.
    """

    rate_source_type: RateSourceType
    """
    The type of input being provided for this CurvePoint, e.g. if it's from a 3M future,
    this would be FUTURE_PX, while if it's from traditional rates fixings, it would be FIXINGS.
    """

    rate_sources: Optional[List[str]]
    """
    The specific rate sources used, e.g. OIS, SOFR or LIBOR; for LENDING_RATE, DeFi or other sources
    used, e.g. CHAINLINK, IPOR or AAVE. For FUNDING_RATE this holds the UUID for the exchange Organization ID.
    """

    reference_assets: Optional[List[UUID]]
    """
    In the case where an implied forward is backed out from market observables, the assets observed.
    """

    mark_prices: Optional[List[float]]
    """
    In the case where a DF is backed out from the implied forward of a reference asset or basket thereof,
    the observed prices that should go into the bootstrapping method.
    """

    rate: float
    """
    The input rate value, if DF not provided.
    """

    discount_factor: float
    """
    The input DF value, if rate not provided.
    """


class YieldCurveDefinition(CamelModel):
    """
    A uniquely-identified set of YC parameters for bootstrapping a YieldCurve.
    """

    yield_curve_id: UUID
    """
    Unique ID for this particular combination of yield curve attributes; note yield curves are
    bootstrapped daily, and so there are going to be many versions over time.
    """

    curve_usage: CurveUsage
    """
    The curve's intended purpose, e.g. for discounting or representing market view on forward rates.
    """

    interpolation_method: InterpolationMethod
    """
    The specific interpolator type used to bootstrap this curve.
    """

    rate_source_type: RateSourceType
    """
    The type of input being provided for this CurvePoint, e.g. if it's from a 3M future,
    this would be FUTURE_PX, while if it's from traditional rates fixings, it would be FIXINGS.
    """

    underlier_asset_id: UUID
    """
    The linked asset for this curve, e.g. for an Ethereum staking curve, this would be ETH (tok.eth.ethereum).
    Note we will be switching to the Exposure UUID instead in a future release (e.g. tok.eth), once the
    reference data is available.
    """

    display_name: str
    """
    Human-readable description of this curve, e.g. OIS (FLAT_FWD)
    """


class YieldCurveAvailability(CamelModel):
    """
    Information about version availability for a given YC definition.
    """

    definition: YieldCurveDefinition
    """
    Description of the particular yield curve parameters that are available to load.
    """

    as_of_times: List[datetime]
    """
    The list of all available as_of_times in the requested window.
    """


class RawYieldCurve(CamelModel):
    """
    A term structure of yield curve inputs. The RAW representation is offered to clients so they
    can either do their own interpolation or for diagnostics.
    """

    points: List[CurvePoint]
    """
    The list of market data observations that went into this raw yield curve, e.g. rates, discount factors
    and futures prices corresponding to various tenors.
    """


class InterpolatedYieldCurve(CamelModel):
    """
    A term structure of rates and discount factors built from a RAW representation. This is the version
    that you should pass in for option valuation purposes, and is suitable for extracting rates and discount
    factors as well as plotting purposes.
    """

    definition: YieldCurveDefinition
    """
    The unique set of parameters used to bootstrap this yield curve.
    """

    durations: List[float]
    """
    Array of all durations along the curve, as year fractions.
    """

    rates: List[float]
    """
    Array of all interpolated rates along the curve.
    """

    discount_factors: List[float]
    """
    Array of all discount factors (DF's) along the curve.
    """


class YieldCurveVersion(CamelModel):
    """
    A single version of a YieldCurveDefinition, inclusive of its raw and interpolated YC content.
    """

    raw: RawYieldCurve
    """
    The raw yield curve input.
    """

    interpolated: InterpolatedYieldCurve
    """
    The bootstrapped yield curve.
    """

    as_of_time: datetime
    """
    The time window, generally UTC midnight, for which we have bootstrapped this yield curve; latest rates / input
    prices as of this time are used.
    """

    build_time: datetime
    """
    The actual time of the build; due to DQ or system issues this might be different from as_of_time.
    """



from typing import List
from uuid import UUID

from serenity_types.utils.serialization import CamelModel


class AssetSectorMappings(CamelModel):
    """
    A single entry in a particular sectory taxonomy mapping a single asset to a sector path.
    """

    asset_id: UUID
    """
    Unique ID for this asset in the Asset Master.
    """

    taxonomy_id: UUID
    """
    Unique ID for the sector taxonomy.
    """

    sector_levels: List[str]
    """
    Fully-qualified sector level to the asset, e.g. something like [”Foo Sector”, “Bar Sub-sector”, “Baz Industry”].
    """


class SectorTaxonomy(CamelModel):
    """
    A single sector taxonomy (hierarchy) known to the system.
    """

    taxonomy_id: UUID
    """
    Unique ID for the sector taxonomy.
    """

    display_name: str
    """
    Short descriptive name, e.g. DATS
    """

    sector_levels: List[str]
    """
    List of names of level types supported, e.g. [”Sector”, “Sub-sector”, “Industry”] or [”Sector”, “Sub-sector”].
    """



from serenity_types.refdata.derivatives import Expiry, ListedDerivative, PayoffType


class Future(ListedDerivative):
    """
    An exchange-listed futures contract.
    """

    expiry: Expiry
    """
    Expiration date and time for this particular term futures contract.
    """

    payoff_type: PayoffType
    """
    Whether the contract tracks the price movement (LINEAR) or its mirror image (INVERSE).
    """


class Perpetual(ListedDerivative):
    """
    An exchange-listed perpetual future, sometimes referred to as a swap.
    """

    payoff_type: PayoffType
    """
    Whether the contract tracks the price movement (LINEAR) or its mirror image (INVERSE).
    """



from datetime import date
from typing import Optional
from uuid import UUID

from serenity_types.utils.serialization import CamelModel
from serenity_types.valuation.core import CashTreatment, MarkTime


class PricingContext(CamelModel):
    """
    Standard settings to use when doing pricing for risk calculation purposes,
    portfolio valuation, etc.. Generally controls which prices to select and
    how to convert those prices into the organization's base currency.
    """

    as_of_date: Optional[date] = None
    """
    The date on which the portfolio was valued; default to latest date.
    """

    mark_time: Optional[MarkTime] = MarkTime.UTC
    """
    The close time convention to use for close-on-close prices in the 24x7 market.
    """

    base_currency_id: Optional[UUID] = None
    """
    The accounting currency to use for valuation, reporting, etc., e.g. fund reports in USD.
    """

    cash_treatment: Optional[CashTreatment] = CashTreatment.FIAT_ONLY
    """
    What to consider to be a cash position, e.g. for NAV calcs.
    """



from datetime import date
from typing import Generic, List, Optional, TypeVar
from uuid import UUID


from serenity_types.utils.serialization import CamelModel

T = TypeVar('T')


class Response(CamelModel, Generic[T]):
    """
    Generic response wrapper that conforms to the Serenity REST standard. The result
    object is the payload in the response and can vary from endpoint to endpoint.
    The other fields are used to convey supporting information like warnings generated
    by the system in response to the request, e.g. if missing data was skipped.

    In some cases we will do some on-the-fly JSON object rewriting to force API
    output into this shape pre-decoding, as the current API is not 100% consistent.
    """

    request_id: UUID
    """
    The correlated request ID for this response.
    """

    as_of_date: Optional[date]
    """
    For reference data and other requests that are loaded for a milestone "as-of" date,
    the date that was loaded. In cases where the inbound request took the default and
    did not pass in an as_of_date this lets the client know the latest date loaded.
    """

    warnings: Optional[List[str]]
    """
    Any warnings from the analytics or data layer to qualify the response. Some API's
    going forward will support a 'strict' mode which does not tolerate warnings.
    """

    result: T
    """
    The response payload; can be anything but typically another Pydantic custom type.
    """



from datetime import date
from typing import List, Optional
from uuid import UUID

from serenity_types.portfolio.core import AssetPosition
from serenity_types.pricing.core import MarkTime
from serenity_types.utils.serialization import CamelModel


class VaRAnalysisRequest(CamelModel):
    """
    A request to compute VaR (Value at Risk) for a single portfolio using a given model.
    """

    as_of_date: Optional[date]
    """
    UTC date for the analysis; default to today.
    """

    horizon_days: Optional[int]
    """
    Loss forecast horizon in days, defaults to 1 day; it is used to scale single day
    VaR by sqrt(horizonDays). Must be positive.
    """

    lookback_period: Optional[int]
    """
    Length of price time series data used to calibrate VaR, measured in days;
    defaults to 365 days.
    """

    mark_time: Optional[MarkTime]
    """
    Enum for close timestamp to use for prices; defaults to NY_EOD.
    """

    base_currency_id: Optional[UUID]
    """
    Base currency to report absolute VaR numbers in, defaults to USD.
    """

    quantiles: Optional[List[float]]
    """
    Loss forecast quantiles used in VaR calculation; defaults to [95, 97.5, 99],
    and must be both unique and > 0 and < 100.
    """

    portfolio: List[AssetPosition]
    """
    The portfolio to analyze.
    """

    model_config_id: UUID
    """
    The specific VaR model ID to use from the repository based on model configuration identifiers from [ModelOps API.
    """


class VaRQuantile(CamelModel):
    """
    Absolute and relative VaR results for a single VaR forecast quantile, e.g. 99%.
    """

    quantile: float
    """
    VaR forecast quantile, e.g. 99 (expressed as percent).
    """

    var_absolute: float
    """
    VaR level (forecasted portfolio loss) in base ccy.
    """

    var_relative: float
    """
    VaR level (forecasted portfolio loss) as a percentage of the baseline portfolio value.
    """


class VaRResult(CamelModel):
    """
    The analysis result for a single day returns the baseline value and the various confidence
    interval values around that baseline.
    """

    run_date: date
    """
    The UTC date for which VaR was computed.
    """

    baseline: float
    """
    The value of the portfolio on the preceding day (runDate - 1).
    """

    quantiles: List[VaRQuantile]
    """
    VaR result object, one for each quantile requested.
    """

    excluded_asset_ids: List[UUID]
    """
    Assets excluded from VaR calculation due to insufficient history.
    """


class VaRAnalysisResult(CamelModel):
    """
    The result of computing VaR on a single portfolio on a single date.
    """

    result: VaRAnalysisRequest
    """
    A single day's result.
    """

    warnings: List[str]
    """
    Any warnings in the analysis to report to the client.
    """


class VaRBacktestRequest(CamelModel):
    """
    A request to backtest VaR (Value at Risk) for a single portfolio over a period using a given model.
    """
    start_date: date
    """
    UTC date for start of analysis run.
    """

    end_date: Optional[date]
    """
    UTC date for end of analysis run, inclusive; default to yesterday.
    """

    lookback_period: Optional[int]
    """
    Length of price time series data used to calibrate VaR, measured in days;
    defaults to 365 days for one year lookback.
    """

    mark_time: Optional[MarkTime]
    """
    Enum for close timestamp to use for prices; defaults to NY_EOD.
    """

    base_currency_id: Optional[UUID]
    """
    Base currency to report absolute VaR numbers in, defaults to USD.
    """

    quantiles: Optional[List[float]]
    """
    Loss forecast quantiles used in VaR calculation; defaults to [95, 97.5, 99],
    and must be both unique and > 0 and < 100.
    """

    portfolio: List[AssetPosition]
    """
    The portfolio to backtest.
    """

    model_config_id: UUID
    """
    The specific VaR model ID to use from the repository based on model configuration identifiers from [ModelOps API.
    """


class VaRBreach(CamelModel):
    """
    Information about a day when the portfolio loss exceeded the VaR forecasted level in a backtest.
    """

    breach_date: date
    """
    Aate when portfolio loss exceeds VaR forecasted level.
    """

    portfolio_loss_absolute: float
    """
    Portfolio loss on breachDate in base ccy.
    """

    portfolio_loss_relative: float
    """
    Portfolio loss on breach_date as a percentage of the baseline value.
    """

    quantiles: List[VaRQuantile]
    """
    VaR levels (absolute & relative) for all breaches this date.
    """


class VaRBacktestResult(CamelModel):
    """
    The result of a full VaR backtest of a portfolio.
    """

    results: List[VaRResult]
    """
    VaR computed results for every day in the backtest.
    """

    breaches: List[VaRBreach]
    """
    Days when the portfolio loss exceeded the VaR forecasted level in a backtest.
    """

    warnings: List[str]
    """
    Any warnings in the backtest to report to the client.
    """
